module Rietveld # :nodoc:

  ##
  # The Role class implements a DCI Role with the Delegator pattern.
  #
  # It tries to be as transparent as possible to the object it delegates to. For
  # instance most BasicObject and Kernel methods are directly forwarded. Though
  # methods concerning methods and / or instance variables return values from
  # the delegator and the object it delegates to.

  class Role < BasicObject
    # Inherit from Basic Object and include Kernel, this is to avoid Rails
    # injecting unwanted methods into our precious Role class
    include ::Kernel

    # The object acting out this role
    attr_accessor :__role_player__

    # Returns the context this role is playing in
    attr_accessor :__context__

    def initialize(role_player, context = nil)
      self.__role_player__ = role_player
      self.__context__     = context
    end

    # Returns the __role_player__ without its role object wrapping it.
    def unwrap
      __role_player__
    end

    # Returns the class of the role itself
    alias_method :role_class, :class

    # Gets a method defined on the role itself
    alias_method :role_method, :method

    # Get an array of methods defined on the role itself
    alias_method :role_methods, :methods

    # Get an array of protected methods defined on the role itself
    alias_method :role_protected_methods, :protected_methods

    # Get an array of public methods defined on the role itself
    alias_method :role_public_methods, :public_methods

    # Get an array of private methods defined on the role itself
    alias_method :role_private_methods, :private_methods

    # Is the other role equal? to this one
    alias_method :role_equal?, :equal?

    # Is the other role eql? to this one
    alias_method :role_eql?, :eql?

    # Get the hash for the role
    alias_method :role_hash, :hash

    # Gets the instance variables defined on the role itself
    alias_method :role_instance_variables,         :instance_variables

    # Get an instance variable defined on the role
    alias_method :role_instance_variable_get,      :instance_variable_get

    # Set an instance variable defined on the role
    alias_method :role_instance_variable_set,      :instance_variable_set

    # Remove an instance variable defined on the role
    alias_method :remove_role_instance_variable,   :remove_instance_variable

    # Is given instance variable defined on the role
    alias_method :role_instance_variable_defined?, :instance_variable_defined?

    # Get the singleton class of the role
    alias_method :role_singleton_class,            :singleton_class

    # Gets the singleton methods defined on the role itself
    alias_method :role_singleton_methods,          :singleton_methods

    # Gets a singleton method defined on the role itself
    alias_method :role_singleton_method,           :singleton_method

    ##--
    ## From BasicObject
    ##++

    # See BasicObject#==, forwarded to __role_player__
    def ==(other)
      __role_player__ == other
    end

    # See BasicObject#equal?, forwarded to __role_player__
    def equal?(other)
      __role_player__.equal?(other)
    end

    # See BasicObject#class, forwarded to __role_player__
    def class
      __role_player__.class
    end

    # See BasicObject#!, forwarded to __role_player__
    def !
      !__role_player__
    end

    # See BasicObject#!=, forwarded to __role_player__
    def !=(other)
      __role_player__ != other
    end

    private

    # See BasicObject#singleton_method_added, forwarded to __role_player__
    def singleton_method_added(sym)
      __role_player__.singleton_method_added(sym)
    end

    # See BasicObject#singleton_method_removed, forwarded to __role_player__
    def singleton_method_removed(sym)
      __role_player__.singleton_method_removed(sym)
    end

    # See BasicObject#singleton_method_undefined, forwarded to __role_player__
    def singleton_method_undefined(sym)
      __role_player__.singleton_method_undefined(sym)
    end

    # If the role itself does not have a given method defined, it will try to
    # delegate it to the role_player
    def method_missing(name, *args, &block)
      if __role_player__.respond_to?(name)
        __role_player__.__send__(name, *args, &block)
      elsif __context__.role?(name)
        __context__.send(name)
      else
        super
      end
    end

    public

    # True if either role object or role_player responds to given method
    def respond_to?(name, include_all = false)
      __role_player__.respond_to?(name, include_all) || super
    end

    # True if either role object or role_player responds to given missing method
    def respond_to_missing?(name, include_private = false)
      __role_player__.respond_to_missing?(name, include_private) || super
    end

    ##--
    ## From Kernel
    ##++

    # See Kernel#nil?, forwarded to __role_player__
    def nil?
      __role_player__.nil?
    end

    # See Kernel#===, forwarded to __role_player__
    def ===(other)
      __role_player__ === other
    end

    # See Kernel#=~, forwarded to __role_player__
    def =~(other)
      __role_player__ =~ other
    end

    # See Kernel#!~, forwarded to __role_player__
    def !~(other)
      __role_player__ !~ other
    end

    # See Kernel#eql?, forwarded to __role_player__
    def eql?(other)
      __role_player__.eql?(other)
    end

    # See Kernel#hash, forwarded to __role_player__
    def hash
      __role_player__.hash
    end

    # See Kernel#<=>, forwarded to __role_player__
    def <=>(other)
      __role_player__ <=> other
    end

    # Returns the class of the __role_player__
    def class
      __role_player__.class
    end

    # Returns the singleton_class of the __role_player__
    def singleton_class
      __role_player__.singleton_class
    end

    # Returns a clone of role object with a clone of the __role_player__
    def clone
      role_clone = super
      role_clone.__role_player__ = __role_player__.clone
      role_clone
    end

    # Returns a dup of __role_player__
    def dup
      __role_player__.dup
    end

    # See Kernel#taint, forwarded to __role_player__
    def taint
      super
      __role_player__.taint
    end

    # See Kernel#tainted?, forwarded to __role_player__
    def tainted?
      super || __role_player__.tainted?
    end

    # See Kernel#untaint, forwarded to __role_player__
    def untaint
      super
      __role_player__.untaint
    end

    # See Kernel#freeze, forwarded to __role_player__
    def freeze
      super
      __role_player__.freeze
    end

    # See Kernel#unfreeze, forwarded to __role_player__
    def unfreeze
      super || __role_player__.frozen?
    end

    # See Kernel#to_s, forwarded to __role_player__
    def to_s
      __role_player__.to_s
    end

    # See Kernel#inspect, forwarded to __role_player__
    #
    # Returns a string in the form of:
    #     #<ExampleRole:0x37d837e #<ExampleObject>>
    def inspect
      "#<#{role_class.name}:0x#{__id__.to_s(16)} #{__role_player__.inspect}>"
    end

    # See Kernel#methods, returns the union of the methods from the role object
    # and __role_player__
    def methods
      super | __role_player__.methods
    end

    # See Kernel#singleton_methods, returns the union of the singleton methods
    # from the role object and __role_player__
    def singleton_methods
      __role_player__.singleton_methods
    end

    # See Kernel#protected_methods, returns the union of the protected_methods
    # from the role object and __role_player__
    def protected_methods
      super | __role_player__.protected_methods
    end

    # See Kernel#private_methods, returns the union of the private_methods
    # from the role object and __role_player__
    def private_methods
      super | __role_player__.private_methods
    end

    # See Kernel#private_methods, returns the union of the private_methods
    # from the role object and __role_player__
    def public_methods
      super | __role_player__.public_methods
    end

    # See Kernel#instance_variables, returns the union of the instance_variables
    # from the role object and __role_player__
    def instance_variables
      super | __role_player__.instance_variables
    end

    # See Kernel#instance_variable_get, prefers to return an instance variable
    # defined on __role_player__, otherwise will lookup instance variables on
    # role
    def instance_variable_get(name)
       __role_player__.instance_variable_get(name) || super
    end

    # See Kernel#instance_variable_set, prefers to set an instance variable
    # on __role_player__, if an instance variable with the same name on the role
    # object is already set, it wil set one on the role.
    def instance_variable_set(name, value)
      if role_instance_variable_defined?(name)
        super
      else
        __role_player__.instance_variable_set(name, value)
      end
    end

    # See Kernel#instance_variable_defined?, returns true if an instance
    # variable is defined on either __role_player__ or the role object
    def instance_variable_defined?(name)
      super || __role_player__.instance_variable_defined?(name)
    end

    # See Kernel#remove_instance_variable, if the instance variable is defined
    # on the role object, it wil remove that one, else it will remove one
    # defined on __role_player__
    def remove_instance_variable(name)
      if role_instance_variable_defined?(name)
        super
      else
        __role_player__.remove_instance_variable(name)
      end
    end

    # See Kernel#instance_of?, forwarded to __role_player__
    def instance_of?(type)
      __role_player__.instance_of?(type)
    end

    # See Kernel#kind_of?, forwarded to __role_player__
    def kind_of?(type)
      __role_player__.kind_of?(type)
    end

    # See Kernel#is_a?, forwarded to __role_player__
    def is_a?(type)
      __role_player__.is_a?(type)
    end

    # See Kernel#method, if given method is defined on role object, it will
    # return that one, otherwise the one from __role_player__
    def method(name)
      super
    rescue
      __role_player__.method(name)
    end

    # See Kernel#public_method, if given public_method is defined on role
    # object, it will return that one, otherwise the one from __role_player__
    def public_method(name)
      super
    rescue
      __role_player__.public_method(name)
    end

    # See Kernel#singleton_method, if given singleton_method is defined on role
    # object, it will return that one, otherwise the one from __role_player__
    def singleton_method(name)
      super
    rescue
      __role_player__.singleton_method(name)
    end
  end
end
